#lang racket

;; Timer%

(require "interfaces.rkt")
(provide Timer%)


;; models a countdown timer.  Counts down from 100 to 0.
;; publishes the count when it changes.
;; accepts a (set-time! n) method to change the timer.

;; a Timer is a (new Timer% [n Number][rate Number])

(define Timer%
  (class* object% (Timer<%>)
    (init-field n) ; Current value of the timer
    (init-field rate)   ; How often does the timer decrement itself
                        ; (in ticks)
    
    (field [time-left rate])     ; Number of ticks til next decrement
    (field [subscribers empty])  ; ListOf<(Number->Void)>

    (field [MAXVAL 100])

    (super-new)

    (define/public (subscribe fn)
      (set! subscribers 
        (cons fn subscribers)))

    (define/public (on-tick)
      (if (zero? time-left)
        (decrement-time)
        (set! time-left (- time-left 1))))

    (define (decrement-time)
      (send this set-time! (if (zero? n) MAXVAL (- n 1))))

    ;; set the time to the new value, reset time-left, and publish
    (define/public (set-time! new-time)
      (set! n new-time)
      (set! time-left rate)
      ;; and publish the new time
      (for-each 
        (lambda (fn) (fn n))
        subscribers))

))

;; and that's all there is!


