#lang racket

(require "interfaces.rkt")
(require "slider.rkt")
(require "display.rkt")
(require 2htdp/image)
(require 2htdp/universe)

(provide World%)

;; A World% consists of a timer and some controllers.  
;; it distributes on-tick to the timer.
;; it distributes on-mouse and add-to-scene to the controllers.
;; it handles keyboard input itself-- factory seems overkill for this.

;; A World is a (new World% [timer Timer%]{[controllers ListOf<Controller<%>>]})
;; {...} denotes optional argument.

(define World%
  (class* object%
    (World<%>)
    
    (init-field timer)                 ; Timer%
    (init-field [controllers empty])   ; ListOf<Controller<%>>

    ;; (Controller<%> -> Void) -> Void
    (define (for-each-controller fn)
      (for-each fn controllers))

    ;; (Controller<%> Y -> Y) Y List<Controller<%>> -> Y
    (define (foldr-controllers fn base)
      (foldr fn base controllers))

    (super-new)

    (define CANVAS-WIDTH 400)
    (define CANVAS-HEIGHT 300)
    (define EMPTY-CANVAS (empty-scene CANVAS-WIDTH CANVAS-HEIGHT))
    (define CENTER-X (/ CANVAS-WIDTH 2))
    (define CENTER-Y (/ CANVAS-HEIGHT 2))

    (define/public (on-tick)
      (send timer on-tick))

    (define/public (on-display)
      (foldr-controllers
       (lambda (controller scene) (send controller add-to-scene scene))
       EMPTY-CANVAS))

    ;; some state to keep track of the last mouse event of any kind
    ;; ie, the  "current" mouse location.
    (field [mouse-x CENTER-X])
    (field [mouse-y CENTER-Y])

    (define/public (on-mouse mx my mev)
      (set! mouse-x mx)
      (set! mouse-y my)
      (for-each-controller 
        (lambda (obj) (send obj on-mouse mx my mev))))

    ;; on "n" create a new slider at current mouse position.  
    ;; "d" creates a new digital display. 
    ;; Other key events are distributed to the controllers.
    ;; Our current controllers don't do anything with key events, but
    ;; other controllers might.
    (define/public (on-key kev)
      (cond
        [(key=? kev "n") 
         (set! controllers
           (cons
             (new Slider% [x mouse-x][y mouse-y][timer timer])
             controllers))]
        [(key=? kev "d") 
         (set! controllers
           (cons
             (new Display% [x mouse-x][y mouse-y][timer timer])
             controllers))]
        [else (for-each-controller
                (lambda (c) (send c on-key kev)))]))
    
    
    ))

